#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <chm_lib.h>

#define BUF_SIZE 16384

typedef struct _list{
	char* file;
	struct _list *next, *prev;
} list;

void free_list( list* lst );
int check_path( char* path );

static void  usage(char *pname)
{
  fprintf(stderr, "Usage: %s chmfile outdir\n", pname);
}

static chm_element( struct chmFile* f, struct chmUnitInfo* ui, void* data )
{
	list* n;
	list** l = (list**)data;
	//Add file to begin of the list
	n = (list*)malloc( sizeof(list) );
	n->next = *l;
	n->prev=NULL;
	if(*l) (*l)->prev=n;
	n->file = strdup(ui->path);
	*l = n;
	return CHM_ENUMERATOR_CONTINUE;
}

list* chm_get_index( struct chmFile* file )
{
	list* l = NULL;
	if( !chm_enumerate( file, CHM_ENUMERATE_ALL, chm_element, &l ) )
		return NULL;
	return l;
}

int main(int argc, char *argv[])
{
  struct chmFile *c;
  char *infname;
  char *outdirname;
  int i,pos;
  int namelen;
  char *rel, *path;
  char dirname[PATH_MAX];
  int err;
  FILE *f;
  list *l,*lt;
  unsigned int offset=0, len=0;
  unsigned char buffer[BUF_SIZE];
  struct chmUnitInfo ui;

  if (3 != argc) {
    usage(argv[0]);
    exit(-1);
  }

  infname = argv[1];
  outdirname = argv[2];
  err = mkdir(outdirname, 0777);
  if (err) {
    if (errno == EEXIST) {
      fprintf(stderr, "Directory %s already exists; you must specify a new directory.\n", outdirname);
    }
    else
      perror("mkdir");
    exit(-1);
  }

  c = chm_open(infname);

  if (!c) {
    perror("chm_openfile");
    exit(-1);
  }

  err = chdir(outdirname);
  if (err != 0) {
    perror("chdir");
    rmdir(outdirname);
    exit(-1);
  }


	l = chm_get_index( c );

	while( l ){
		lt = l;
		l = l->next;
	}

/*	while( lt ){
		printf("%s\n",lt->file);
		lt = lt->prev;
	}
*/
	while( lt ){
    if ( lt->file[0] == '/') {
      namelen = strlen(lt->file);
      if (namelen == 1){
		  lt=lt->prev;
		  continue; // skip the slash 
	  }
      rel = lt->file + 1;
      if ( lt->file[namelen-1] == '/') {
	  	strncpy(dirname, rel, namelen-2);
		dirname[namelen-2] = 0;
		check_path( dirname );
 		err = mkdir(dirname, 0777);
		if (err != 0 && err != EEXIST) {
		  free_list( l );
		  fprintf(stderr, "mkdir failed on %s\n", dirname);
		  perror("mkdir");
		  exit(-1);
 	  	}
	  }else {
		path = strdup(rel);
		check_path( path );
		free(path);
			
		f = fopen(rel, "wb");
		if (!f) {
			free_list( l );
			fprintf(stderr, "Couldn't open %s\n", rel);
			perror("fopen");
			exit(-1);
		}
	
		if( chm_resolve_object( c, lt->file, &ui ) != CHM_RESOLVE_SUCCESS){
			chm_close( c );
			free_list( l );
			exit(-1);
		}
		
		offset = len = 0;
		while( offset < ui.length ){
			if( ui.length - offset < BUF_SIZE ) len =  ui.length - offset;
			else len = BUF_SIZE;
		
		    len =  (unsigned int)chm_retrieve_object( c, &ui, buffer, offset, len );
	    	offset += len;
			fwrite( buffer, 1, len, f);
		}
		fclose( f );
  	} 
  }
  lt=lt->prev;
  }
  
  chm_close(c);
  free_list(l);
  return 0;
}

void free_list( list* lst )
{
  list *lt, *l = lst;

  while( l ){
  if(l->file != NULL) free(l->file);
	  lt = l;
	  l=l->next;
	  free( lt );
  }
}

int check_path( char* path )
{
	int pos,beg,err;
	struct stat st;
	  
	pos = strlen( path )-1;
	while( pos >= 0 && path[pos] != '/') pos--;
	if( pos > 0 ){
		//check dir path
		path[pos] = 0;
		if( stat(path,&st) == -1 && errno == ENOENT ){
			//create path dirs
			beg = check_path( path )+1;
 			err = mkdir(path, 0777);
			if (err != 0 && err != EEXIST) {
		  		fprintf(stderr, "mkdir: failed on %s\n", path);
				perror("mkdir");
				exit(-1);
	 	  	}			
		}
		path[pos] = '/';
	}
	return pos;
}
